module CustomAttributes
  class FieldType
    class_attribute :type_name
    class_attribute :multiple_supported

    # Defines if this type
    self.multiple_supported = false

    # Pretty name of the field
    self.type_name = nil

    def name
      self.class.type_name
    end

    def label
      "label_#{name}"
    end

    def self.available_types
      descendants
    end

    def self.find(name)
      "CustomAttributes::#{name}FieldType".constantize.instance
    rescue NameError
      nil
    end

    # Validate a CustomValue according to the type roles. This method can  be
    # overridden by field types if the way bulk handling should happen changes
    # (e.g. List type)
    def validate_custom_value(custom_value)
      # 1. Wrap in Array so we can pass Array of values and single values, reject empty values
      values = Array.wrap(custom_value.value).reject { |value| value.to_s == '' }
      # 2. Validate each value
      errors = values.map do |value|
        validate_single_value(custom_value.custom_field, value, custom_value.customizable)
      end
      errors.flatten.uniq
    end

    # Override this in subclass to validate custom values
    def validate_single_value(_custom_field, _value, _customizable = nil)
      []
    end

    # Overide this in subclass to validate custom fields
    def validate_custom_field(_custom_field)
      []
    end

    def after_save_custom_value(custom_field, custom_value); end

    def before_custom_field_save(custom_field); end

    # Prepares and sets CustomFieldValues value.
    def set_custom_field_value(_custom_field, _custom_field_value, value)
      if value.is_a?(Array)
        value = value.map(&:to_s).reject { |v| v == '' }.uniq
        value << '' if value.empty?
      else
        value = value.to_s
      end

      value
    end

    # Cast the value of an existing CustomValue
    def cast_custom_value(custom_value)
      cast_value(custom_value.custom_field, custom_value.value, custom_value.customizable)
    end

    # Cast the value according to field type rules
    def cast_value(custom_field, value, customizable = nil)
      if value.blank?
        nil
      elsif value.is_a?(Array)
        casted = value.map do |v|
          cast_single_value(custom_field, v, customizable)
        end
        casted.compact.sort
      else
        cast_single_value(custom_field, value, customizable)
      end
    end

    def cast_single_value(_custom_field, value, _customizable = nil)
      value.to_s
    end

    # Checks if a keyword is in the possible values and returns the possible value if found
    # If there are no possible values, returns the keyword
    def value_from_keyword(custom_field, keyword, object)
      possible_values_options = possible_values_options(custom_field, object)
      if possible_values_options.present?
        parse_keyword(custom_field, keyword) do |k|
          if v = possible_values_options.detect { |text, _id| k.casecmp(text) == 0 }
            if v.is_a?(Array)
              v.last
            else
              v
            end
          end
        end
      else
        keyword
      end
    end

    # Parses keyword, allows comma delimited values
    def parse_keyword(custom_field, keyword)
      separator = Regexp.escape ','
      keyword = keyword.to_s

      if custom_field.multiple?
        values = []
        until keyword.empty?
          k = keyword.dup
          loop do
            if value = yield(k.strip)
              values << value
              break
            elsif k.slice!(/#{separator}([^#{separator}]*)\Z/).nil?
              break
            end
          end
          keyword.slice!(/\A#{Regexp.escape k}#{separator}?/)
        end
        values
      else
        yield keyword.strip
      end
    end
    protected :parse_keyword

    # Use to get the possible values of a CustomValue
    def possible_custom_value_options(custom_value)
      possible_values_options(custom_value.custom_field, custom_value.customizable)
    end

    # Override this in subclass to specify the possible values for the field with this type
    def possible_values_options(_custom_field, _object = nil)
      []
    end

    # Returns the HTML Tag to edit this field format.
    def edit_tag(view, tag_id, tag_name, custom_value, options = {})
      view.text_field_tag(tag_name, custom_value.value, options.merge(id: tag_id))
    end

  end
end
